#! perl -w
#
# demo_win32_printers.pl
#
# Author:  Bill Odom
#
# Put Win32::Printers through its paces.
#

## Pragmas

use strict;


## Modules

# Add the lib directory in the home directory of this
# script to the lib path.
#
use FindBin;
use File::Spec;
use lib File::Spec->catdir($FindBin::Bin, 'lib');

use Win32;

# Win32::Printers doesn't export anything by default, but the :ALL
# tag will pull in everything from it.
#
use Win32::Printers qw(:ALL);


## Variables

my %printer_drivers;
my %printers;
my %printer;
my $ph;
my $answer;
my $printer_connection;


## Subroutines

# sub separator
#
# A little convenience routine that prints a separator line.
#
sub separator
  {
    print '=' x 75, "\n";
  }

# sub outcome
#
# This keeps me from having to stick in the ternary operator
# all over the place.
#
sub outcome
  {
    return (scalar(shift) ? 'succeeded.' : 'failed: ' . Win32::FormatMessage(Win32::GetLastError) );
  }

## Main

#
# Start by finding the printer driver directory on the system.
# This is a good thing to know when you start copying files to the system
# for use by AddPrinterDriver.
#

print "\nPrinter driver directory:\n", GetPrinterDriverDirectory(), "\n\n";

separator;


%printer_drivers = Win32::Printers::EnumPrinterDrivers;

if (scalar(keys %printer_drivers))
  {
    print "\nPrinter drivers installed on this system:\n\n";

    for my $d (sort keys %printer_drivers)
      {
        print $printer_drivers{$d}->{Name}, "\n";

        for my $k (sort keys %{ $printer_drivers{$d} })
          {
            # next unless ($printer_drivers{$d}->{$k});
            next if ($k eq 'Name');
            printf("  %-16s: %s\n", $k, $printer_drivers{$d}->{$k} ? $printer_drivers{$d}->{$k} : '<null>' );
          }
        print "\n";
      }
  }
else
  {
    print "\nNo printer drivers are installed on this system.\n\n";
  }

separator;

#
# Print a list of all the locally-installed printers, if there are any.
#

%printers = EnumPrinters();

if (scalar(keys %printers))
  {
    print "\nPrinters installed on this system:\n\n";

    # EnumPrinters returns a hash of hashes.  The outer hash is
    # keyed on the printer name, and the inner hash keys are the
    # names of the printer attributes.
    #
    for my $p (sort keys %printers)
      {
        print $printers{$p}->{PrinterName}, ' (', $printers{$p}->{PortName}, ")\n";
    
        for my $k ( sort keys %{ $printers{$p} } )
          {
            next if ($k eq 'PrinterName') or ($k eq 'PortName');
            printf("  %-25s: %s\n", $k, $printers{$p}->{$k} );
          }
        print "\n";
      }

    separator;

    # Pick a printer from the hash of printer information.
    #
    %printer = %{ $printers{ (sort keys %printers)[0] } };

    # Try to open it.
    #
    print "\nAttempting to open [$printer{PrinterName}]...\n" ,
          outcome( $ph = OpenPrinter( $printer{PrinterName} ) ) ,
          "\n";

    # If we were able to open it, try to set it as the default, first with
    # SetPrinter, then with SetDefaultPrinter.  Then close it.
    #
    if ($ph)
      {
        if (Win32::IsWin95)
          {
            print "\nAttempting to set [$printer{PrinterName}]\n",
                  "as the default using SetPrinter...\n",
                  outcome( SetPrinter($ph,
                    { %printer , Attributes => PRINTER_ATTRIBUTE_DEFAULT } ) ) ,
                  "\n";
          }

        print "\nAttempting to close [$printer{PrinterName}]...\n" ,
              outcome( ClosePrinter($ph) ) ,
              "\n";
      }

    # Now that all the opening and closing is done, try to set the default
    # using a more direct approach.
    #
    if (Win32::IsWin95)
      {
        print "\nAttempting to set [$printer{PrinterName}]\n",
              "as the default using SetDefaultPrinter...\n",
              outcome( SetDefaultPrinter( $printer{PrinterName} ) ) ,
              "\n";
      }
  
  }
else
  {
    print "\nNo local printers are installed.\n";
  }

print "\n";

separator;


#
# For the remainder of this program to work, fill in the parameters
# with values meaningful on your network.  The $printer_connection
# variable immediately below is a good place to start.
#

$printer_connection = '\\\\wnodomsvr01\\stl6n450hp1';

# The AddPrinterConnection and DeletePrinterConnection API calls
# only work on NT, so these perl wrappers only work on NT as well.
#
# But take heart -- I'm experimenting with a couple
# of higher-level calls that know how to add drivers and printers
# simultaneously (for both NT and 95), but they're not ready
# to be unleashed on an unsuspecting public.
#
if (Win32::IsWinNT)
  {
    print "\nAttempting to add a printer connection...\n",
          outcome( AddPrinterConnection( $printer_connection ) ),
          "\n";
    
    print "\nAttempting to delete a printer connection...\n",
          outcome( DeletePrinterConnection( $printer_connection ) ),
          "\n";
  }

# And these two -- you'll definitely need to change these, unless the printer
# you happen to be adding is really an Apple LaserWriter on a server
# called SOMESERVER behind a share called SOMEPRINTER.  Hey, it could happen.
#
# Also, these don't currently work on NT (since I originally only needed them
# for 95), but I hope to fix that soon.  See the note before the [if (Win32::IsWinNT)]
# block immediately above.
#
if (Win32::IsWin95)
  {
    print "\nAttempting to add a printer driver...\n",
          outcome( 
            AddPrinterDriver(
              Version         => 1024 ,
              Environment     => 'Windows 4.0',
              Name            => 'Apple LaserWriter' ,
              DataFile        => 'APPLE230.SPD' ,
              ConfigFile      => 'PSCRIPT.DRV' ,
              DriverPath      => 'PSCRIPT.DRV' ,
              HelpFile        => 'PSCRIPT.HLP' ,
              MonitorName     => 'PostScript Language Monitor' ,
              DependentFiles  =>
                [qw(
                  APPLE230.SPD PSCRIPT.DRV PSCRIPT.HLP PSCRIPT.INI TESTPS.TXT
                  APPLE380.SPD FONTS.MFM ICONLIB.DLL PSMON.DLL
                )] ,
            )
          ) ,
          "\n";

    print "Attempting to add a printer...\n",
          outcome( 
            $ph = AddPrinter(
              PrinterName => 'LaserWriter',
              PortName    => $printer_connection,
              DriverName  => 'Apple LaserWriter',
              Comment     => 'The LaserWriter printer on some server',
            )
          ) ,
          "\n";

    if ($ph)
      {
        print "Attempting to delete a printer by handle...\n",
              outcome( DeletePrinter( $ph ) ) ,
              "\n";

        ClosePrinter($ph);
      }
  }

print "\n";

separator;

print "\nDone.\n";

# end demo_win32_printers.pl

__END__

