#! perl -w
#
# Win32::Printers.pm
#
# Author: Bill Odom
#
# The early, early beginnings of a module to manage printers under Windows.
#
# Currently, this should be considered more an example than a production module.
# It can certainly do useful work under both NT and 95, and it's a worthwhile
# demonstration of using Windows API calls from Perl -- but you should expect it
# to change significantly over the next few weeks.
#

## Module Interface

package Win32::Printers;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
require Exporter;
@ISA = qw(Exporter);

# Don't export anything by default.  If the caller wants everything,
# specifying :ALL in the use statement will do the trick.
#
@EXPORT = ();

@EXPORT_OK = qw(

    EnumPrinters
    OpenPrinter
    ClosePrinter

    DeletePrinter
    DeletePrinterByName
    DeletePrinterByPort
    DeletePrinterConnection

    SetPrinter
    SetDefaultPrinter

    GetPrinterDriverDirectory

    AddPrinterDriver
    AddPrinter
    AddPrinterConnection

    PRINTER_ATTRIBUTE_QUEUED
    PRINTER_ATTRIBUTE_DIRECT
    PRINTER_ATTRIBUTE_DEFAULT
    PRINTER_ATTRIBUTE_SHARED
    PRINTER_ATTRIBUTE_NETWORK
    PRINTER_ATTRIBUTE_HIDDEN
    PRINTER_ATTRIBUTE_LOCAL

    PRINTER_ATTRIBUTE_ENABLE_DEVQ
    PRINTER_ATTRIBUTE_KEEPPRINTEDJOBS
    PRINTER_ATTRIBUTE_DO_COMPLETE_FIRST

    PRINTER_ATTRIBUTE_WORK_OFFLINE
    PRINTER_ATTRIBUTE_ENABLE_BIDI
    PRINTER_ATTRIBUTE_RAW_ONLY
    PRINTER_ATTRIBUTE_PUBLISHED

    PRINTER_ENUM_DEFAULT
    PRINTER_ENUM_LOCAL
    PRINTER_ENUM_CONNECTIONS
    PRINTER_ENUM_FAVORITE
    PRINTER_ENUM_NAME
    PRINTER_ENUM_REMOTE
    PRINTER_ENUM_SHARED
    PRINTER_ENUM_NETWORK

    PRINTER_ENUM_EXPAND
    PRINTER_ENUM_CONTAINER

    PRINTER_ENUM_ICONMASK
    PRINTER_ENUM_ICON1
    PRINTER_ENUM_ICON2
    PRINTER_ENUM_ICON3
    PRINTER_ENUM_ICON4
    PRINTER_ENUM_ICON5
    PRINTER_ENUM_ICON6
    PRINTER_ENUM_ICON7
    PRINTER_ENUM_ICON8
  );

# Right now, there's only one export tag, but as the number of functions
# handled by this module grows, so will this hash.
#
%EXPORT_TAGS =
  (
    ALL => [ @EXPORT_OK ] ,
  );

$VERSION = '0.01';  # er... is there anything before "alpha" ?


## Modules

use Win32::API;


## Constants

#
# This list will be expanded, but for now I'm keeping it small.
#

sub NULL () { 0 }

sub PRINTER_ATTRIBUTE_QUEUED            () { 0x00000001 }
sub PRINTER_ATTRIBUTE_DIRECT            () { 0x00000002 }
sub PRINTER_ATTRIBUTE_DEFAULT           () { 0x00000004 }
sub PRINTER_ATTRIBUTE_SHARED            () { 0x00000008 }
sub PRINTER_ATTRIBUTE_NETWORK           () { 0x00000010 }
sub PRINTER_ATTRIBUTE_HIDDEN            () { 0x00000020 }
sub PRINTER_ATTRIBUTE_LOCAL             () { 0x00000040 }

sub PRINTER_ATTRIBUTE_ENABLE_DEVQ       () { 0x00000080 }
sub PRINTER_ATTRIBUTE_KEEPPRINTEDJOBS   () { 0x00000100 }
sub PRINTER_ATTRIBUTE_DO_COMPLETE_FIRST () { 0x00000200 }

sub PRINTER_ATTRIBUTE_WORK_OFFLINE      () { 0x00000400 }
sub PRINTER_ATTRIBUTE_ENABLE_BIDI       () { 0x00000800 }
sub PRINTER_ATTRIBUTE_RAW_ONLY          () { 0x00001000 }
sub PRINTER_ATTRIBUTE_PUBLISHED         () { 0x00002000 }

sub PRINTER_ENUM_DEFAULT                () { 0x00000001 }
sub PRINTER_ENUM_LOCAL                  () { 0x00000002 }
sub PRINTER_ENUM_CONNECTIONS            () { 0x00000004 }
sub PRINTER_ENUM_FAVORITE               () { 0x00000004 }
sub PRINTER_ENUM_NAME                   () { 0x00000008 }
sub PRINTER_ENUM_REMOTE                 () { 0x00000010 }
sub PRINTER_ENUM_SHARED                 () { 0x00000020 }
sub PRINTER_ENUM_NETWORK                () { 0x00000040 }

sub PRINTER_ENUM_EXPAND                 () { 0x00004000 }
sub PRINTER_ENUM_CONTAINER              () { 0x00008000 }

sub PRINTER_ENUM_ICONMASK               () { 0x00ff0000 }
sub PRINTER_ENUM_ICON1                  () { 0x00010000 }
sub PRINTER_ENUM_ICON2                  () { 0x00020000 }
sub PRINTER_ENUM_ICON3                  () { 0x00040000 }
sub PRINTER_ENUM_ICON4                  () { 0x00080000 }
sub PRINTER_ENUM_ICON5                  () { 0x00100000 }
sub PRINTER_ENUM_ICON6                  () { 0x00200000 }
sub PRINTER_ENUM_ICON7                  () { 0x00400000 }
sub PRINTER_ENUM_ICON8                  () { 0x00800000 }

## Variables

#
# All file-level lexicals have been moved into enclosing
# scopes around the API wrapper subroutines.
#

## Subroutines

#
# The following two subroutines, null_terminate and lop, are
# pretty handy to have around, and will probably be moved to
# a more general module in the future.  They're not exported,
# and I don't expect that to change.
#

# sub null_terminate
#
# A convenience routine that ensures each of its arguments is
# null-terminated.  Although it modifies the arguments in-place,
# the list of arguments is also returned.
#
sub null_terminate
  {
    for my $s (@_) { $s = $s . "\0" unless substr($s,-1) eq "\0"; }

    return (wantarray ? @_ : $_[0]);
  }
# end sub null_terminate


# sub lop
#
# Another convenience routine that lops off everything after the
# last null character in each specified string.  Modifications are
# made in place, and the list of strings is also returned.
#
sub lop
  {
    for my $s (@_) { $s =~ s/\0.*$//; }

    return (wantarray ? @_ : $_[0]);
  }
# end sub lop


BEGIN # EnumPrinters enclosing scope
  {
    my $EnumPrinters;

    # sub EnumPrinters
    #
    sub EnumPrinters
      {
        $EnumPrinters ||= Win32::API->new(
          'winspool.drv' ,
          'EnumPrinters' ,
          [qw( L P L P L P P )] ,
          'N'
        ) or return;

        # Parameters - none

        # Variables
        #
        my $bytes_needed;
        my $pl_bytes_needed;
        my $structs_copied;
        my $return;
        my $sizeof_pi5_struct;
        my $info_structs;
        my %pi5structs;

        # Reserve space for the two LONGs.
        #
        $bytes_needed   = ' ' x 4;
        $structs_copied = ' ' x 4;

        # Call EnumPrinters once to find out how many bytes we actually need.
        # This call will "fail" (its return value will be zero), but
        # $bytes_needed (when unpacked) will tell us how much memory is
        # required on the *next* call.
        #
        $EnumPrinters->Call(
           PRINTER_ENUM_LOCAL ,
           NULL ,
           5 ,
           NULL ,
           NULL ,
           $bytes_needed ,
           $structs_copied ,
          );

        # We use this a couple of times, so save the unpacked value for later.
        #
        $pl_bytes_needed = unpack('L', $bytes_needed);

        # Now that we know how much room we need, we can allocate space
        # for a whole lot of PRINTER_INFO_5 structures, then call EnumPrinters
        # again to get it.
        #
        $info_structs = ' ' x $pl_bytes_needed;

        # If *this* call fails, return undef to the caller.
        #
        $EnumPrinters->Call(
           PRINTER_ENUM_LOCAL ,
           NULL ,
           5 ,
           $info_structs ,
           $pl_bytes_needed ,
           $bytes_needed ,
           $structs_copied ,
          )
        or return;

        # Since we no longer need the packed representations of these
        # two values, just overwrite them with the unpacked values.
        #
        ($bytes_needed, $structs_copied) = unpack('LL', $bytes_needed . $structs_copied);

        # Figure out how big a PRINTER_INFO_5 struct is, so we can do some math.
        #
        $sizeof_pi5_struct = 4 * 5;  # 2 LPTSTRs and 3 DWORDS

        # Chop up the block of PRINTER_INFO_5 structs that EnumPrinters returned to us,
        # turning the raw data into a nice, convenient hash.
        #
        for (my $i = 0; $i < $structs_copied; $i++)  # <-- Look! I remember C!
          {
            my %this_pi5;

            @this_pi5{
               'PrinterName' ,
               'PortName'    ,
               'Attributes'  ,
               'DeviceNotSelectedTimeout' ,
               'TransmissionRetryTimeout'
              } = unpack('ppL3', substr($info_structs, $i * $sizeof_pi5_struct, $sizeof_pi5_struct) );

            # Stick the reference to this hash into the bigger hash.
            #
            $pi5structs{ lc $this_pi5{PrinterName} } = \%this_pi5;
          }

        return %pi5structs;
      }
    # end sub EnumPrinters
  }
# end EnumPrinters enclosing scope


BEGIN # EnumPrinterDrivers enclosing scope
  {
    my $EnumPrinterDrivers;

    sub EnumPrinterDrivers
      {
        $EnumPrinterDrivers ||= Win32::API->new(
          'winspool.drv',
          'EnumPrinterDrivers',
          [qw( P P L P L P P )],
          'N'
        ) or return;

        my %parameters =
          (
            Name        => "" ,
            Environment => ( Win32::IsWinNT() ? 'Windows NT x86' : 'Windows 4.0' ) ,
            @_
          );

        my $buffer = '';
        my $bytes_needed     = ' ' x 4;
        my $drivers_returned = ' ' x 4;
        my $pl_drivers_returned = 0;
        my $sizeof_di3_struct;
        my %di3structs;

        null_terminate( $parameters{Name}, $parameters{Environment} );

        $EnumPrinterDrivers->Call(
            $parameters{Name},
            $parameters{Environment},
            3,
            $buffer,
            length($buffer),
            $bytes_needed,
            $drivers_returned,
          );

        $buffer = (' ' x unpack('L', $bytes_needed));

        $EnumPrinterDrivers->Call(
            $parameters{Name},
            $parameters{Environment},
            3,
            $buffer,
            length($buffer),
            $bytes_needed,
            $drivers_returned,
         ) or return;

        $pl_drivers_returned = unpack('L', $drivers_returned);

        $sizeof_di3_struct = 4 + (4 * 9); # one DWORD and 9 LPTSTRs

        for (my $i = 0; $i < $pl_drivers_returned; $i++)
          {
            my %this_di3;

            @this_di3{qw(
              Version
              Name
              Environment
              DriverPath
              DataFile
              ConfigFile
              HelpFile
              DependentFiles
              MonitorName
              DefaultDataType
            )} = unpack('Lp9', substr($buffer, $i * $sizeof_di3_struct, $sizeof_di3_struct) );

            # Stick the reference to this hash into the bigger hash.
            #
            $di3structs{ lc $this_di3{Name} } = \%this_di3;
          }

        return %di3structs;
      }
  }
# end EnumPrinterDrivers enclosing scope


BEGIN # OpenPrinter enclosing scope
  {
    my $OpenPrinter;

    # sub OpenPrinter
    #
    # Given a printer name, this function opens the printer
    # corresponding to this name, then returns a "handle" to
    # the printer.  This handle can then be used to change
    # the status of the printer, delete the printer, close
    # the printer, etc.
    #
    sub OpenPrinter
      {
        $OpenPrinter ||= Win32::API->new(
          'winspool.drv',
          'OpenPrinter',
          [qw( P P P )]
          , 'N'
        ) or return;

        # Parameters
        #
        my $printer_name = shift;

        # Variables
        #
        my $printer_handle = ' ' x 4; # sizeof DWORD

        null_terminate($printer_name);

        return
          $OpenPrinter->Call( $printer_name, $printer_handle, NULL)
          ? $printer_handle
          : undef
          ;
      }
    # end sub OpenPrinter
  }
# end OpenPrinter enclosing scope


BEGIN # ClosePrinter enclosing scope
  {
    my $ClosePrinter;

    # sub ClosePrinter
    #
    # Closes the printer corresponding to the specified printer handle.
    #
    sub ClosePrinter
      {
        $ClosePrinter ||= Win32::API->new(
          'winspool.drv',
          'ClosePrinter',
          [qw( L )],
          'N'
        ) or return;

        # Parameters
        #
        my $printer_handle = shift;

        # Begin

        return $ClosePrinter->Call( unpack('L', $printer_handle) );
      }
    # end sub ClosePrinter
  }
# end ClosePrinter enclosing scope


BEGIN # SetPrinter enclosing scope
  {
    my $SetPrinter;

    # sub SetPrinter
    #
    # Given a printer handle and a reference to a hash, this function
    # attempts to set the status and attributes of the specified printer.
    #
    sub SetPrinter
      {
        $SetPrinter ||= Win32::API->new(
          'winspool.drv',
          'SetPrinter',
          [qw( L L P L )],
          'N'
        ) or return;

        # According to the MS docs, the SetPrinter() API call works under both NT
        # and 95, but this doesn't match my experience. Also, I've hard-wired
        # this wrapper to use PRINTER_INFO_5 structures, which are _also_
        # supposed to work under NT, but don't. For those reasons, I'm forcing
        # the function to return under NT.  I hope to change this soon.
        #
        return if Win32::IsWinNT;

        # Parameters
        #
        my $printer_handle = shift;
        my $printer_info   = shift;

        # Variables
        #
        my $printer_info_packed;

        # Begin

        # Create a PRINTER_INFO_5 structure from the specified hash.
        #
        $printer_info_packed = pack( 'ppL3',
          $printer_info->{PrinterName},
          $printer_info->{PortName},
          $printer_info->{Attributes},
          $printer_info->{DeviceNotSelectedTimeout},
          $printer_info->{TransmissionRetryTimeout},
        );

        return $SetPrinter->Call( unpack('L', $printer_handle), 5, $printer_info_packed, NULL );
      }
    # end sub SetPrinter
  }
# end SetPrinter enclosing scope


BEGIN # DeletePrinter enclosing scope
  {
    my $DeletePrinter;

    # sub DeletePrinter
    #
    # Deletes the printer corresponding to the specified printer handle.
    #
    sub DeletePrinter
      {
        my $DeletePrinter ||= Win32::API->new(
          'winspool.drv',
          'DeletePrinter',
          [qw( L )],
          'N'
        ) or return;

        # As with SetPrinter, the docs claim this works under NT,
        # but just not for me, I suppose.  Here's the cop-out.
        #
        return if Win32::IsWinNT;

        # Parameters
        #
        my $printer_handle = shift;

        # Begin

        return $DeletePrinter->Call( unpack('L', $printer_handle) );
      }
    # end sub DeletePrinter
  }
# end DeletePrinter enclosing scope


# sub DeletePrinterByName
#
# Attempts to delete the printer corresponding to the specified name.
# Case is not significant.
#
sub DeletePrinterByName
  {
    # Parameters
    #
    my $printer_name = lc shift;

    # Variables
    #
    my $printer_handle;
    my $success;
    my %printer_info;

    # Begin

    # Get the list of printers into a hash.
    #
    %printer_info = EnumPrinters();

    # An error-checking block, to avoid nested if's.
    #
    BLOCK:
      {
        $success = 0;

        $printer_info{$printer_name}
          or last;

        $printer_handle = OpenPrinter( $printer_info{$printer_name}->{PrinterName} )
          or last;

        $success = DeletePrinter($printer_handle);

        ClosePrinter($printer_handle)
          or warn "Unable to close printer";
      }

    return $success;
  }
# sub DeletePrinterByName


# sub DeletePrinterByPort
#
# Attempts to delete the printer corresponding to the specified port.
# Case is not significant.
#
sub DeletePrinterByPort
  {
    # Parameters
    #
    my $port_name = lc shift;

    # Variables
    #
    my $printer_name;
    my $printer_handle;
    my $success;
    my %printer_info;

    # Begin

    # Get the list of printers into a hash.
    #
    %printer_info = EnumPrinters();

    # An error-checking block, to avoid nested if's.
    #
    BLOCK:
      {
        $success = 0;

        # Find the printer name associated with this port.
        #
        for my $p (values %printer_info)
          {
            if ( lc($p->{PortName}) eq $port_name)
              {
                $printer_name = lc $p->{PrinterName};
                last;
              }
          }

        last unless $printer_name;

        $printer_handle = OpenPrinter( $printer_info{$printer_name}->{PrinterName} )
          or last;

        $success = DeletePrinter($printer_handle);

        ClosePrinter($printer_handle)
          or warn "Unable to close printer";
      }

    return $success;
  }
# sub DeletePrinterByPort


BEGIN # DeletePrinterConnection enclosing scope
  {
    my $DeletePrinterConnection;

    # sub DeletePrinterConnection
    #
    sub DeletePrinterConnection
      {
        # Hey, here's a switch -- this one only works under NT.
        #
        return unless Win32::IsWinNT;

        $DeletePrinterConnection ||= Win32::API->new(
          'winspool.drv',
          'DeletePrinterConnection',
          [qw( P )],
          'N'
        ) or return;

        my $printer_name = shift;

        null_terminate($printer_name);

        return $DeletePrinterConnection->Call($printer_name);
      }
    # end sub DeletePrinterConnection
  }
# end DeletePrinterConnection enclosing scope


# sub SetDefaultPrinter
#
# A shortcut for setting the default printer.
#
sub SetDefaultPrinter
  {
    # Parameters
    #
    my $printer_name = lc shift;

    # Variables
    #
    my $printer_handle;
    my $success;
    my %printer_info;

    # Begin

    # Get the list of printers into a hash.
    #
    %printer_info = EnumPrinters();

    # An error-checking block, to avoid nested if's.
    #
    BLOCK:
      {
        $success = 0;

        $printer_info{$printer_name}
          or last;

        $printer_handle = OpenPrinter( $printer_info{$printer_name}->{PrinterName} )
          or last;

        my %printer = %{ $printer_info{$printer_name} };

        $printer{Attributes} = ($printer{Attributes} | PRINTER_ATTRIBUTE_DEFAULT);

        $success = SetPrinter( $printer_handle, \%printer );

        ClosePrinter($printer_handle)
          or warn "Unable to close printer";
      }

    return $success;
  }
# end sub SetDefaultPrinter


BEGIN # GetPrinterDriverDirectory enclosing scope
  {
    my $GetPrinterDriverDirectory;

    # sub GetPrinterDriverDirectory
    #
    sub GetPrinterDriverDirectory
      {
        $GetPrinterDriverDirectory ||= Win32::API->new(
          'winspool.drv',
          'GetPrinterDriverDirectory',
          [qw( P P L P L P )],
          'N'
        ) or return;

        # Parameters - none

        # Variables
        #
        my $driver_directory;
        my $success;
        my $bytes_needed;

        # Begin

        # Start with a buffer that's too small, so GetPrinterDriverDirectory()
        # will tell us how much room it actually needs.
        #
        $driver_directory = '';
        $bytes_needed     = ' ' x 4;
        $GetPrinterDriverDirectory->Call(
          NULL,
          NULL,
          1,
          $driver_directory,
          length($driver_directory),
          $bytes_needed
        );

        # Now use the $bytes_needed value to create a buffer of the proper
        # size and call GetPrinterDriverDirectory again.
        # If this call fails, just return an undef.
        #
        $driver_directory = ' ' x unpack('L', $bytes_needed);
        $GetPrinterDriverDirectory->Call(
          NULL,
          NULL,
          1,
          $driver_directory,
          length($driver_directory),
          $bytes_needed
          )
        or return;

        # Remove everything after the trailing null and return
        # the driver directory.
        #
        lop($driver_directory);

        return $driver_directory;

      }
    # end sub GetPrinterDriverDirectory
  }
# end GetPrinterDriverDirectory enclosing scope


BEGIN # AddPrinterDriver enclosing scope
  {
    my $AddPrinterDriver;

    # sub AddPrinterDriver
    #
    sub AddPrinterDriver
      {
        # I haven't been able to get this working under NT yet.
        #
        return if Win32::IsWinNT;

        $AddPrinterDriver ||= Win32::API->new(
          'winspool.drv',
          'AddPrinterDriver',
          [qw( P L P )],
          'N'
        ) or return;

        # Parameters (named parameters, with some default values)
        #
        # Note: The MS docs say that the version on 95/98 should be
        # zero, but most of the drivers I've used actually specify 1024,
        # so that's what I'm using here.
        #
        my %parameters =
          (
            Environment     => ( Win32::IsWinNT() ? 'Windows NT x86' : 'Windows 4.0' ) ,
            Version         => ( Win32::IsWinNT() ? 2                : 1024          ) ,
            DefaultDataType => 'EMF' ,
            @_
          );

        # Variables
        #
        my $di3;
        my $dependent_files;
        my $driver_path;

        $driver_path = GetPrinterDriverDirectory();

        # Convert the dependent files string into the proper format.
        # If the parameter is an array ref, then join the array elements
        # together with nulls.
        #
        if (ref($parameters{DependentFiles}) eq 'ARRAY')
          {
            $dependent_files = join("\0", @{ $parameters{DependentFiles} });
          }
        else
          {
            ($dependent_files = $parameters{DependentFiles} ) =~ s/;/\0/g;
          }

        # Double null-terminate the list of dependent files for AddPrinterDriver()
        # API call.
        #
        null_terminate($dependent_files); # make sure the first one is there...
        $dependent_files .= "\0";         # ...then add another one.

        # Prepend the driver directory path to the four main files.
        #
        for my $file (@parameters{'DriverPath', 'DataFile', 'ConfigFile', 'HelpFile'})
          {
            $file = $driver_path . "\\" . $file;
          }

        # Null-terminate the strings.
        #
        for my $string ( @parameters{ qw(
          Name Environment DataFile
          ConfigFile HelpFile MonitorName DefaultDataType
          DriverPath DataFile ConfigFile HelpFile) } )
          {
            null_terminate($string);
          }

        # Build a DRIVER_INFO_3 structure for passing to AddPrinterDriver.
        #
        $di3 = pack( 'Lp9',
          $parameters{Version},
          $parameters{Name},
          $parameters{Environment},
          $driver_path,
          $parameters{DataFile},
          $parameters{ConfigFile},
          $parameters{HelpFile},
          $dependent_files,
          $parameters{MonitorName},
          $parameters{DefaultDataType}
        );

        return $AddPrinterDriver->Call( NULL, 3, $di3 );
      }
    # end sub AddPrinterDriver
  }
# end AddPrinterDriver enclosing scope


BEGIN # AddPrinter enclosing scope
  {
    my $AddPrinter;

    # sub AddPrinter
    #
    # Printer names are limited to 32 characters (under 9x, at least).
    # Exceeding this limit will cause the AddPrinter() API call to fail.
    # Be careful.
    #
    sub AddPrinter
      {
        $AddPrinter ||= Win32::API->new(
          'winspool.drv',
          'AddPrinter',
          [qw( P L P )],
          'N'
        ) or return;

        # Parameters
        #
        my %parameters = (
          PrintProcessor => 'WinPrint',
          DataType       => 'EMF',
          Comment        => '',
          Location       => '',
          @_
        );

        # Variables
        #
        my $pi2;
        my $printer_handle;

        # Begin

        for my $string ( @parameters{
            qw( PrinterName PortName PrintProcessor DataType Comment Location ) } )
          {
            null_terminate($string);
          }

        # Build a PRINTER_INFO_2 structure.
        #
        $pi2 = pack('L p L p3 L3 p2 L10',
          NULL, # Server name
          $parameters{PrinterName},
          NULL, # Share name
          $parameters{PortName},
          $parameters{DriverName},
          $parameters{Comment},
          NULL, # Location
          NULL, # DevMode
          NULL, # SepFile
          $parameters{PrintProcessor},
          $parameters{DataType},
          NULL, # Parameters
          NULL, # SecurityDescriptor
          NULL, # Attributes
          NULL, # Priority
          NULL, # DefaultPriority
          NULL, # StartTime
          NULL, # UntilTime
          NULL, # Status
          NULL, # Jobs
          NULL, # AveragePPM
        );

        $printer_handle = $AddPrinter->Call( NULL, 2, $pi2 );

        return ( $printer_handle ? pack('L', $printer_handle) : () );

      }
    # end sub AddPrinter
  }
# end AddPrinter enclosing scope


BEGIN # AddPrinterConnection enclosing scope
  {
    my $AddPrinterConnection;

    # sub AddPrinterConnection
    #
    sub AddPrinterConnection
      {
        return unless Win32::IsWinNT;

        $AddPrinterConnection ||= Win32::API->new(
          'winspool.drv',
          'AddPrinterConnection',
          [qw( P )],
          'N'
        ) or return;

        my $printer_name = shift;

        null_terminate($printer_name);

        return $AddPrinterConnection->Call($printer_name);
      }
    # end sub AddPrinterConnection
  }
# end AddPrinterConnection enclosing scope

#
# Why is GetProfileString in here, you ask?
#
# Well, many of the print spooler functions in the Win32 API, such as SetPrinter,
# don't work as within NT.  So, for example, to set the default printer under
# NT, you have to call GetProfileString and WriteProfileString against the
# proper INI file entries -- which actually pass through to the Registry.
#
# So GetProfileString is here because I'm planning to implement the NT
# functionality soon, and this is the beginning of that effort.  It's not
# exported, and probably won't be, so there's no danger of namespace pollution.
#


BEGIN # GetProfileString enclosing scope
  {
    my $GetProfileString;

    # sub GetProfileString
    #
    sub GetProfileString
      {
        $GetProfileString ||= Win32::API->new(
          'kernel32',
          'GetProfileString',
          [qw( P P P P L )],
          'N'
        ) or return;

        # Parameters
        #
        my $section_name   = shift;
        my $key_name       = shift;
        my $default_string = shift;

        # Variables
        #
        my $destination_buffer = ' ' x 1024;
        my $destination_buffer_size = '';
        my @destination_buffer;

        # Begin

        for my $s ($section_name, $key_name, $default_string)
          {
            null_terminate($s) if $s;
          }

        while (1)
          {
            # Attempt to grab the requested information with the
            # current destination buffer.
            #
            $destination_buffer_size = $GetProfileString->Call(
              $section_name || NULL ,
              $key_name || NULL ,
              $default_string || NULL,
              $destination_buffer ,
              length($destination_buffer),
            );

            # Get out now, unless the returned destination buffer size is
            # within two characters of the actual destination buffer size.
            # If it is, then we probably ran out of room...
            #
            last unless $destination_buffer_size >= (length($destination_buffer)-2);

            # ...so tack on another K and try again.
            #
            $destination_buffer = ' ' x (length($destination_buffer) + 1024);
          }

        # Carve out the relevant part of the destination buffer.
        #
        $destination_buffer = substr($destination_buffer, 0, $destination_buffer_size);

        # Remove all trailing nulls.
        #
        $destination_buffer =~ s/\0+$//;

        # Split the destination buffer into individual elements by using
        # the null byte as a delimiter.
        #
        @destination_buffer = split("\0", $destination_buffer);

        return (wantarray ? @destination_buffer : $destination_buffer[0] );
      }
    # end sub GetProfileString
  }
# end GetProfileString enclosing scope


## Main

1;

#
# I'm not even going to think of adding any POD until the
# functionality has grown and the interface has settled down.
# Once that happens, I promise to document everything
# meticulously.  No, really -- I mean it.
#

__END__
